from trypnv import Machine, PluginStateMachine
from trypnv.nvim import HasNvim
from trypnv.machine import ModularMachine, Transitions

from proteome.nvim import NvimFacade
from proteome.logging import Logging

from tryp import List


class ProteomeComponent(ModularMachine, HasNvim, Logging):

    def __init__(self, name: str, vim: NvimFacade) -> None:
        Machine.__init__(self, name)
        HasNvim.__init__(self, vim)


class ProteomeState(PluginStateMachine, HasNvim, Logging):

    def __init__(self, vim: NvimFacade, plugins: List[str]) -> None:
        HasNvim.__init__(self, vim)
        PluginStateMachine.__init__(self, 'proteome', plugins)


class ProteomeTransitions(Transitions, HasNvim):

    def __init__(self, machine, *a, **kw):
        Transitions.__init__(self, machine, *a, **kw)
        HasNvim.__init__(self, machine.vim)

__all__ = ('ProteomeComponent', 'ProteomeState', 'ProteomeTransitions')



from typing import Callable
from pathlib import Path

from trypnv.record import Record, dfield

from tryp import Map, Maybe, Just
from tryp.transformer import Transformer

from proteome.logging import Logging
from proteome.project import Project
from proteome.git import RepoT, ProjectRepoAdapter, RepoState


class HistoryState(Record):
    repos = dfield(Map())
    browse = dfield(Map())


class History(Logging):

    def __init__(self, base: Path, state: HistoryState=HistoryState()) -> None:
        self.base = base
        self.state = state

    @property
    def repos(self):
        return self.state.repos

    def adapter(self, project: Project):
        git_dir = self.base / project.fqn
        return ProjectRepoAdapter(project, Just(git_dir))

    def state_for(self, project: Project):
        return self.repos.get(project)

    def at(self, project: Project, f: Callable[[RepoT], RepoT]):
        return self.adapter(project)\
            .t(self.state_for(project))\
            .map(f)\
            .map(lambda r: self.repos + (project, r.state))\
            .map(lambda r: self.state.set(repos=r))

    def repo(self, project: Project):
        return self.adapter(project).repo(self.state_for(project))


class HistoryT(Transformer[History]):

    def pure(self, h: Maybe[HistoryState]) -> History:  # type: ignore
        new_state = h | self.state
        return History(self.val.base, new_state)

    @property
    def state(self):
        return self.val.state

__all__ = ('HistoryState', 'History', 'HistoryT')



import tryp.logging
from trypnv.logging import trypnv_logger

from tryp.lazy import lazy


log = proteome_root_logger = trypnv_logger('proteome')


def proteome_logger(name: str):
    return proteome_root_logger.getChild(name)


class Logging(tryp.logging.Logging):

    @lazy
    def _log(self) -> tryp.logging.Logger:  # type: ignore
        return proteome_logger(self.__class__.__name__)

__all__ = ('proteome_logger', 'Logging')



from proteome.plugins.core.message import (
    StageI, StageII, StageIII, StageIV, Add, RemoveByIdent, Create, Next, Prev,
    SetProject, SetProjectIdent, SetProjectIndex, SwitchRoot, Save, Added,
    Removed, ProjectChanged, BufEnter, Initialized, MainAdded, Show,
    AddByParams, CloneRepo)
from proteome.plugins.core.main import Plugin

__all__ = ('StageI', 'StageII', 'StageIII', 'StageIV', 'Add', 'RemoveByIdent',
           'Create', 'Next', 'Prev', 'SetProject', 'SetProjectIdent',
           'SetProjectIndex', 'SwitchRoot', 'Save', 'Added', 'Removed',
           'ProjectChanged', 'BufEnter', 'Initialized', 'MainAdded', 'Show',
           'AddByParams', 'CloneRepo')



from tryp import Path

from proteome.project import Project
from proteome.git import Git


class HistoryGit(Git):

    def __init__(self, base: Path, vim) -> None:
        self.base = base
        super(HistoryGit, self).__init__(vim)

    def pre_args(self, project: Project):
        d = str(project.root)
        h = str(self._history_dir(project))
        return [
            '--git-dir',
            h,
            '--work-tree',
            d,
        ]

    def _history_dir(self, project: Project):
        return self.base / project.fqn

__all__ = ('HistoryGit',)



import re
from pathlib import Path

from fn import F, _

from tryp import List, Map, Empty
from tryp.lazy import lazy

from trypnv.machine import handle, may_handle, Error, Info
from trypnv.process import JobClient

from proteome.state import ProteomeComponent, ProteomeTransitions
from proteome.project import Project, mkpath
from proteome.git import Git
from proteome.plugins.core.message import (
    StageI, StageIV, Add, RemoveByIdent, Create, Next, Prev,
    SetProject, SetProjectIdent, SetProjectIndex, SwitchRoot, Added,
    Removed, ProjectChanged, BufEnter, Initialized, MainAdded, Show,
    AddByParams, CloneRepo)


class Plugin(ProteomeComponent):

    class Transitions(ProteomeTransitions):

        def _no_such_ident(self, ident: str, params):
            return 'no project found matching \'{}\''.format(ident)

        @may_handle(StageI)
        def stage_1(self):
            main = self.data.analyzer(self.vim).main
            return Add(main), MainAdded().pub  # type: ignore

        @may_handle(StageIV)
        def stage_4(self):
            return BufEnter(self.vim.buffer).pub, Initialized().pub

        @may_handle(Initialized)
        def initialized(self):
            return self.data.set(initialized=True), SwitchRoot(False)

        @may_handle(MainAdded)
        def main_added(self):
            self.data.main.foreach(self._setup_main)

        def _setup_main(self, pro: Project):
            self.vim.set_pvar('main_name', pro.name)
            self.vim.set_pvar('main_ident', pro.ident)
            self.vim.set_pvar('main_type', pro.tpe | 'none')
            self.vim.set_pvar('main_types', pro.all_types)

        @may_handle(AddByParams)
        def add_by_params(self):
            params = Map(self.msg.params)
            ident = self.msg.ident
            return (
                (params.get('root') /
                 mkpath //
                 F(self.data.loader.from_params, ident, params=params))
                .or_else(
                    self.data.loader.by_ident(ident)
                    .or_else(self.data.loader.resolve_ident(
                        ident, params, self.data.main_type))
                ) /
                Add |
                Error(self._no_such_ident(ident, params))
            )

        @may_handle(Add)
        def add(self):
            if self.msg.project not in self.data:
                return (self.data.add(self.msg.project),
                        Added(self.msg.project).pub)

        @may_handle(Added)
        def added(self):
            self.vim.set_pvar('added_project', self.msg.project.json)
            self.vim.set_pvar('projects', self.data.projects.json)
            self.vim.pautocmd('Added')
            if self.data.initialized:
                return SetProjectIndex(-1)

        # TODO switch project if removed current
        @handle(RemoveByIdent)
        def remove_by_ident(self):
            return self.data.project(self.msg.ident)\
                .map(lambda a: (self.data - a, Removed(a).pub))

        @may_handle(Removed)
        def removed(self):
            self.vim.set_pvar('projects', self.data.projects.json)
            self.vim.pautocmd('Removed')
            return Info('Removed project {}'.format(self.msg.project.ident))

        @may_handle(Create)
        def create(self):
            return self.data + Project.of(self.msg.name, Path(self.msg.root))

        @may_handle(Show)
        def show(self):
            lines = self.data.projects.show(List.wrap(self.msg.names))
            header = List('Projects:')  # type: List[str]
            return Info('\n'.join(header + lines))

        @may_handle(SetProject)
        def set_project(self):
            if isinstance(self.msg.ident, str):
                if self.msg.ident in self.data:
                    return SetProjectIdent(self.msg.ident)
                elif self.msg.ident.isdigit():
                    return SetProjectIndex(int(self.msg.ident))
                else:
                    err = '\'{}\' is not a valid project identifier'
                    return Error(err.format(self.msg.ident))

        @may_handle(SetProjectIndex)
        def set_project_index(self):
            if self.msg.index < self.data.project_count:
                return self.data.set_index(self.msg.index), SwitchRoot()

        @handle(SetProjectIdent)
        def set_project_ident(self):
            return self.data.set_index_by_ident(self.msg.ident)\
                .map(lambda a: (a, SwitchRoot()))

        @handle(SwitchRoot)
        def switch_root(self):
            def go(pro: Project):
                self.vim.switch_root(pro.root)  # type: ignore
                pc = ProjectChanged(pro)
                info = 'switched root to {}'
                return (pc, Info(info.format(pro.ident))) if self.msg.notify else pc
            if self.data.initialized:
                return self.data.current.map(go)
            else:
                return Empty()

        @may_handle(ProjectChanged)
        def project_changed(self):
            self.vim.set_pvar('active', self.msg.project.json)

        @may_handle(Next)
        def next(self):
            return self.data.inc(1), SwitchRoot()

        @may_handle(Prev)
        def prev(self):
            return self.data.inc(-1), SwitchRoot()

        @may_handle(Error)
        def error(self):
            self.log.error(self.msg.message)

        @may_handle(Info)
        def info(self):
            self.log.info(self.msg.message)

        # TODO make configurable (destination dir)
        @handle(CloneRepo)
        def clone_repo(self):
            uri = self.msg.uri
            name = self.msg.params.get('name')\
                .or_else(self._clone_repo_name(uri))
            url = self._clone_url(uri)
            return (
                self.data.main_clone_dir.ap2(name, _ / _)
                .to_either('invalid parameter: {}'.format(uri)) /
                F(self._clone_repo, url)
            )

        def _clone_url(self, uri):
            return uri if uri.startswith('http') else self._github_url(uri)

        def _github_url(self, uri):
            return 'https://github.com/{}'.format(uri)

        def _clone_repo_name(self, uri):
            return (
                List.wrap(uri.split('/'))
                .lift(-1) /
                F(re.sub, '\.git$', '')
            )

        @property
        def cloner(self) -> Git:
            return self._cloner  # type: ignore

        @lazy
        def _cloner(self):
            return Git(self.vim)

        async def _clone_repo(self, url: str, target):
            ident = '/'.join(str(target).split('/')[-2:])
            client = JobClient(cwd=Path.home(), name=self.name)
            res = await self.cloner.clone(client, url, target)
            return res.either(AddByParams(ident, {}).pub,
                              'failed to clone {} to {}'.format(url, target))

__all__ = ('Plugin',)



from pathlib import Path
from datetime import datetime

from fn import _, F

from dulwich.object_store import tree_lookup_path

from tryp.lazy import lazy
from tryp import Map, __, Just, Empty, may, List, Maybe, Right
from tryp.util.numeric import try_convert_int
from tryp.async import gather_sync_flat

from trypnv.machine import (may_handle, message, handle, IO, RunTask, Info)
from trypnv.machine import Error
from trypnv.record import field, dfield, Record, lazy_list_field, maybe_field
from trypnv.nvim import ScratchBuilder, ScratchBuffer

from proteome.state import ProteomeComponent, ProteomeTransitions
from proteome.plugins.core import Save, StageIV
from proteome.logging import Logging
from proteome.project import Project
from proteome.git import Repo, CommitInfo
from proteome.plugins.history.messages import (HistoryPrev, HistoryNext,
                                               HistoryStatus, HistoryLog,
                                               HistoryBrowse,
                                               HistoryBrowseInput,
                                               HistorySwitch, Redraw,
                                               QuitBrowse, Commit,
                                               HistoryBufferNext,
                                               HistoryBufferPrev, HistoryPick,
                                               HistoryRevert,
                                               HistoryFileBrowse,
                                               HistorySwitchFile,
                                               CommitCurrent)
from proteome.plugins.history.data import History, HistoryT, HistoryState
from proteome.plugins.history.process import HistoryGit
from proteome.plugins.history.patch import Patch


class BrowseState(Record):
    repo = field(Repo)
    current = field(int)
    commits = lazy_list_field()
    buffer = field(ScratchBuffer)
    selected = dfield(0)
    path = maybe_field(Path)

Init = message('Init')


class BrowseMachine(ProteomeComponent):
    _data_type = BrowseState

    class Transitions(ProteomeTransitions):

        @property
        def buffer(self):
            return self.data.buffer.proxy

        @lazy
        def content(self):
            sel = self.data.selected
            return List.wrap(enumerate(self.data.commits[:sel + 20]))\
                .flat_smap(lambda i, a: a.browse_format(i == sel))

        @property
        def selected_commit(self):
            return self.data.commits.lift(self.data.selected)

        @property
        def selected_id(self):
            return self.selected_commit / _.hex

        def _create_mappings(self):
            List.wrap('jksprq').foreach(self._create_mapping)
            self._create_mapping('<cr>', to='%CR%')

        def _create_mapping(self, keyseq, mode='n', to=None):
            to_seq = Maybe(to) | keyseq
            raw = self.buffer.buffer
            cmd = ':ProHistoryBrowseInput {}<cr>'.format(to_seq)
            raw.nmap(keyseq, cmd)

        def _configure_appearance(self):
            self.buffer.set_options('filetype', 'diff')
            self.buffer.set_options('syntax', 'diff')
            self.buffer.window.set_optionb('cursorline', True)
            self.vim.doautocmd('FileType')
            sy = self.buffer.syntax
            id_fmt = '[a-f0-9]\+'
            sy.match('Commit', '^[* ]\s\+{}'.format(id_fmt),
                     contains='Star,Sha')
            sy.match('Star', '^*', 'contained')
            sy.match('Sha', id_fmt, 'contained')
            sy.link('Star', 'Title')
            sy.link('Sha', 'Identifier')

        @may_handle(Init)
        def browse_init(self):
            self._create_mappings()
            self._configure_appearance()
            return Redraw()

        @may_handle(Redraw)
        def redraw(self):
            self.buffer.set_content(self.content)
            self.vim.cursor(self.data.selected + 1, 1)
            self.vim.feedkeys('zz')

        @handle(HistoryBrowseInput)
        def input(self):
            handlers = Map({
                'j': self._down,
                'k': self._up,
                '%CR%': self._switch,
                's': self._switch,
                'p': self._pick,
                'r': self._revert,
                'q': self._close_tab,
            })
            return handlers.get(self.msg.keyseq).flat_map(lambda f: f())

        def _down(self):
            return self._select_diff(1)

        def _up(self):
            return self._select_diff(-1)

        @may
        def _select_diff(self, diff):
            index = self.data.selected + diff
            if index >= 0 and self.data.commits.min_length(index + 1):
                return self.data.set(selected=index), Redraw()

        @may
        def _switch(self):
            q = QuitBrowse(self.buffer)
            switch = (
                self.data.path //
                (lambda path: self.selected_id / F(HistorySwitchFile, path)) |
                HistorySwitch(self.data.selected)
            )
            return (switch.pub, q, q.pub)

        @may_handle(QuitBrowse)
        def quit(self):
            if self.msg.buffer == self.buffer:
                self._close_tab()

        @may
        def _close_tab(self):
            self.buffer.tab.close()

        def _pick(self):
            return self._pick_commit(HistoryPick)

        def _revert(self):
            return self._pick_commit(HistoryRevert)

        @may
        def _pick_commit(self, tpe):
            q = QuitBrowse(self.buffer)
            return (tpe(self.data.selected).pub, q, q.pub)


class Browse(Logging):

    def __init__(self, state: BrowseState, vim) -> None:
        self.state = state
        self.machine = BrowseMachine('history_browse', vim)

    @property
    def buffer(self):
        return self.state.buffer

    @property
    def repo(self):
        return self.state.repo

    def run(self):
        return self.send(Init())

    def send(self, msg):
        result = self.machine.loop_process(self.state, msg)
        self.state = result.data
        return result.pub


class Plugin(ProteomeComponent):
    failed_pick_err = 'couldn\'t revert commit'

    @lazy
    def base(self):
        return self.vim.pdir('history_base').get_or_else(Path('/dev/null'))

    @lazy
    def executor(self):
        return HistoryGit(self.base, self.vim)

    class Transitions(ProteomeTransitions):

        @property
        def executor(self):
            return self.machine.executor

        def _with_sub(self, state):
            return self.data.with_sub_state(self.name, state)

        @property
        def _with_browse(self):
            return F(self._with_sub) << self.state.setter('browse')

        @property
        def state(self):
            return self.data.sub_state(self.name, HistoryState)

        @lazy
        def history(self):
            return History(self.machine.base, state=self.state)

        @lazy
        def history_t(self):
            return HistoryT(self.history)

        @property
        def all_projects_history(self):
            return self.pflags.all_projects_history

        @lazy
        def projects(self):
            return self.data.history_projects(self.all_projects_history)

        @property
        def current(self):
            return self.data.current

        def _all_projects(self, f):
            new_state = self.projects.fold_left(self.history_t)(f).state
            return self._with_sub(new_state)

        def _with_repos(self, f):
            g = lambda hist, pro: hist / __.at(pro, _ / f)
            return self._all_projects(g)

        def _with_repo(self, pro, f):
            new_state = (self.history_t / __.at(pro, f)).state
            return self._with_sub(new_state)

        def _with_current_repo(self, f):
            return self.current.map(lambda a: self._with_repo(a, f))

        def _repo_ro(self, project: Project):
            return self.history.repo(project)

        @property
        def _current_repo_ro(self):
            return self.current // self._repo_ro

        @property
        def _repos_ro(self):
            return self.projects // self._repo_ro

        @may_handle(StageIV)
        def stage_4(self):
            ''' initialize repository states '''
            return self._with_repos(lambda a: Just(a.state))

        def _switch(self, f):
            def notify(repo):
                self.vim.reload_windows()
                repo.current_commit_info\
                    .map(_.num_since)\
                    .foreach(self.log.info)
                return repo
            return self._with_current_repo(_ / f % notify)

        # FIXME save first?
        @handle(HistoryPrev)
        def prev(self):
            return self._switch(__.prev())

        @handle(HistoryNext)
        def next(self):
            return self._switch(__.next())

        @handle(HistorySwitch)
        def switch(self):
            return try_convert_int(self.msg.index)\
                .map(__.select)\
                .flat_map(self._switch)

        @handle(HistorySwitchFile)
        def switch_file(self):
            return (
                self._current_repo_ro /
                __.checkout_file(self.msg.id, self.msg.path) /
                __.map(lambda a: CommitCurrent()) /
                RunTask
            )

        @may_handle(HistoryBufferPrev)
        def history_buffer_prev(self):
            pass

        @may_handle(HistoryStatus)
        def history_status(self):
            def log_status(repo):
                state = 'dirty' if repo.status else 'clean'
                return Info('history repo {}'.format(state)).pub
            self._with_current_repo(_ / log_status)

        @may_handle(HistoryLog)
        def history_log(self):
            self._current_repo_ro / _.log_formatted % self.vim.multi_line_info

        def _build_browse(self, repo, commits, path: Maybe[Path]):
            relpath = path // repo.relpath
            return ScratchBuilder().build.unsafe_perform_io(self.vim)\
                .leffect(self._io_error)\
                .map(lambda a: BrowseState(repo=repo, current=0,
                                           commits=commits, buffer=a,
                                           path=relpath))

        def _browse(self, f, path: Maybe[Path]=Empty()):
            commits = lambda r: (r, f(r))
            return (
                (self._current_repo_ro / commits)
                .flat_smap(F(self._build_browse, path=path)) /
                F(self._add_browse)
            )

        @handle(HistoryBrowse)
        def history_browse(self):
            return self._browse(_.history_info)

        @handle(HistoryFileBrowse)
        def history_file_browse(self):
            path = Path(
                self.vim.buffer.name if self.msg.path == '' else self.msg.path)
            return self._current_repo_ro // __.relpath(path) / (
                lambda p: self._browse(__.file_history_info(p),
                                       path=Just(p))
            ) | Right(
                Error('current file \'{}\' not in current repo'.format(path)))

        def _io_error(self, exc):
            self.log.caught_exception('nvim io', exc)

        @may_handle(Save)
        def save(self):
            return Commit()

        @property
        def _timestamp(self):
            return datetime.now().isoformat()

        @handle(HistoryBrowseInput)
        def history_browse_input(self):
            return self._current_browse.map(__.send(self.msg))

        @property
        def _current_browse(self):
            return self._browse_for_buffer(self.vim.buffer)

        def _browse_for_buffer(self, buffer):
            return self.state.browse\
                .find(_.buffer.buffer == buffer)\
                .map(_[1])

        @handle(QuitBrowse)
        def quit_browse(self):
            return self._browse_for_buffer(self.msg.buffer)\
                .map(self._remove_browse)

        def _add_browse(self, state: BrowseState):
            browse = Browse(state, self.vim)
            return (
                self._with_browse(self.state.browse + (browse.repo, browse)),
                IO(browse.run)
            )

        def _remove_browse(self, target: Browse):
            return self._with_browse(self.state.browse - target.repo)

        @handle(HistoryPick)
        def pick(self):
            return self._pick_commit\
                .flat_map(lambda a: self._pick_patch(a[0], a[1]))

        @handle(HistoryRevert)
        def revert(self):
            return self._pick_commit\
                .map(lambda a: self._pick_revert(a[0], a[1]))

        @property
        def _pick_commit(self):
            index = try_convert_int(self.msg.index)
            lifter = self._current_repo_ro / _.history_info / _.lift
            return index.ap(lifter).flatten.product(self.current)

        def _pick_patch(self, commit: CommitInfo, project: Project):
            patch = (
                commit.diff /
                _.revert //
                _.patch
            )
            apply = lambda pat: self._apply_patch(project, pat, commit)
            return patch.map(apply)

        @may
        def _check_pick_status(self, commit, result):
            if result.success:
                self.vim.reload_windows()
                self.log.info('picked {}'.format(commit.num_since))
            else:
                return Error(Plugin.failed_pick_err).pub

        async def _apply_patch(self, project, patch, commit):
            executor = Patch(self.vim)
            result = await executor.patch(project, patch)
            return self._check_pick_status(commit, result)

        async def _pick_revert(self, commit: CommitInfo, project: Project):
            result = await self.executor.revert(project, commit)
            ret = self._check_pick_status(commit, result)
            if not result.success:
                await self.executor.revert_abort(project)
            return ret

        async def _add_commit_coro(self, data):
            project, repo = data
            return await repo.add_commit_all(project, self.executor,
                                             self._timestamp)

        # TODO handle broken repo
        # TODO allow specifying target
        @may_handle(Commit)
        async def commit(self):
            wanted = self.msg.projects
            filt = (lambda p: not wanted or wanted.exists(p.match_ident))
            candidates = self.projects.filter(filt).flat_pair(self._repo_ro)
            results = await gather_sync_flat(candidates, self._add_commit_coro)
            new_repos = results\
                .fold_map(self.state.repos, lambda s: (s.project, s))
            return Just(self._with_sub(self.state.set(repos=new_repos)))

        @may_handle(CommitCurrent)
        def commit_current(self):
            return Commit(*(self.data.current / _.name).to_list)

__all__ = ('Plugin',)



from trypnv.machine import message

Commit = message('Commit', varargs='projects')
CommitCurrent = message('CommitCurrent')
HistorySwitch = message('HistorySwitch', 'index')
HistorySwitchFile = message('HistorySwitchFile', 'path', 'id')
HistoryPrev = message('HistoryPrev')
HistoryNext = message('HistoryNext')
HistoryBufferPrev = message('HistoryBufferPrev')
HistoryBufferNext = message('HistoryBufferNext')
HistoryStatus = message('HistoryStatus')
HistoryLog = message('HistoryLog')
HistoryBrowse = message('HistoryBrowse')
HistoryFileBrowse = message('HistoryBrowse', opt_fields=(('path', ''),))
HistoryBrowseInput = message('HistoryBrowseInput', 'keyseq')
HistoryPick = message('HistoryPick', 'index')
HistoryRevert = message('HistoryRevert', 'index')
Redraw = message('Redraw')
QuitBrowse = message('QuitBrowse', 'buffer')
ExecPick = message('ExecPick', 'commit', 'executor', 'status')
RevertAbort = message('ExecPick', 'project', 'executor', 'status')



from proteome.plugins.history.main import Plugin
from proteome.plugins.history.data import History, HistoryT
from proteome.plugins.history.process import HistoryGit

__all__ = ('Plugin', 'History', 'HistoryT', 'HistoryGit')



from fn import F, _

from trypnv.machine import may_handle

from proteome.state import ProteomeComponent, ProteomeTransitions
from proteome.plugins.core import MainAdded, StageIII


class Plugin(ProteomeComponent):

    class Transitions(ProteomeTransitions):

        @property
        def _project_dir(self):
            return self.vim.pvar('config_project_dir')\
                .get_or_else('project')

        @property
        def _project_after_dir(self):
            return self.vim.pvar('config_project_after_dir')\
                .get_or_else('project_after')

        def _runtime(self, project, base):
            run = F(self.vim.runtime) << F('{}/{}'.format, base)
            project.all_types.foreach(run)
            project.tpe.map(_ + '/' + project.name).foreach(run)
            run('all/*')

        def _runtime_before(self, project):
            return self._runtime(project, self._project_dir)

        def _runtime_after(self, project):
            return self._runtime(project, self._project_after_dir)

        @may_handle(MainAdded)
        def before(self):
            self.data.current.foreach(self._runtime_before)

        @may_handle(StageIII)
        def stage_3(self):
            self.data.current.foreach(self._runtime_after)

__all__ = ('Plugin',)



from pathlib import Path

from tryp.test import temp_dir

import tryp.test
from tryp import Just

from proteome.project import Project


class Spec(tryp.test.Spec):

    def setup(self):
        super().setup()
        self.temp_projects = Path(temp_dir('projects'))
        self.history_base = Path(temp_dir('history'))

    def mk_project(self, name, tpe):
        root = temp_dir(str(self.temp_projects / tpe / name))
        return Project.of(name, Path(root), tpe=Just(tpe))

    def object_files(self, pro: Project):
        objdir = self.history_base / pro.fqn / 'objects'
        return list((objdir).iterdir())

__all__ = ('Spec',)



from proteome.project import Project
from trypnv import ProcessExecutor, Job

from tryp import Just


class Patch(ProcessExecutor):  # type: ignore

    def patch(self, project: Project, diff: str):
        args = ['-p1', '-r-']
        job = Job(
            client=project.job_client,
            exe='patch',
            args=args,
            loop=self.loop,
            pipe_in=Just(diff))
        return self.run(job)

__all__ = ('Patch',)
